-------------------------------------------------------------------
--===============================================================--
--=								=--
--=			RampageAI 				=--
--=				by Faithful			=--
--=					(Robert Grimm)		=--
--=								=--
--=		http://gorgerush.net/faithful/RampageAI/	=--
--=								=--
--===============================================================--
--								 --
-- This custom AI was written completely from scratch, with any	 --
-- code "borrowed" from other scripts or people each labeled	 --
-- with the credit that is due.					 --
--								 --
--===============================================================--
--								 --
-- This script was written in an effort to be as fast,		 --
-- efficient, and customizable as possible. Feel free to browse	 --
-- the source of it and, hopefully, learn something. Note that	 --
-- since I personally am against using such tactics as 3rd	 --
-- party programs to add ctrl-1 thru 9, you will find no such	 --
-- "features" as part of the base script.			 --
--								 --
--===============================================================--
--								 --
-- I highly recommend that customizations be contained within	 --
-- the "Custom.lua" file, such that when newer versions of this	 --
-- script come out, you will be able to quickly and easily	 --
-- transfer your customizations into the newer version.	Doing	 --
-- so will also allow you to easily share your customizations.	 --
--								 --
-- And yes, if you feel you must, you can add customizations	 --
-- of the type I am against (stated above).			 --
--								 --
--===============================================================--
--								 --
--		Enjoy!						 --
--								 --
--===============================================================--
-------------------------------------------------------------------


--[[

This file is part of RampageAI.

RampageAI is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

RampageAI is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RampageAI; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

--]]

---------------------
-- Script Location --
---------------------

-- Set this variable to the directory where your script is stored
--	(request by Fallen~Angel)
-- NOTE: If commented out, Rampage will auto-detect its location


-- ScriptLocation = "./AI/USER_AI/"

-------------------------
-- Require Replacement --
-------------------------

-- Save the old Require
RAI_Require = require

-- Function to check if a file exists
function FileExists(filename)
	-- Open the file, to test if it exists
	local file = io.open(filename)
	if file then
		file:close()
		return true
	else
		return false
	end
end

-- Make a new require function
require = function(script)
	-- Check if the script exists without adding ScriptLocation
	if FileExists(script) then
		return RAI_Require(script)

	-- Check if the script exists in ScriptLocation
	elseif ScriptLocation and FileExists(ScriptLocation .. script) then
		return RAI_Require(ScriptLocation .. script)

	-- Check if the script exists in ./AI/USER_AI/
	elseif FileExists("./AI/USER_AI/" .. script) then
		-- Should we set ScriptLocation?
		if not ScriptLocation then
			ScriptLocation = "./AI/USER_AI/"
		end

		return RAI_Require("./AI/USER_AI/" .. script)

	-- Check if the script exists in ./AI/
	elseif FileExists("./AI/" .. script) then
		-- Should we set ScriptLocation?
		if not ScriptLocation then
			ScriptLocation = "./AI/"
		end

		return RAI_Require("./AI/" .. script)

	end

	-- Since we can't find it, just pass thru to probably get an error
	return RAI_Require(script)
end

---------------------
-- Support Scripts --
---------------------

-- Before we load any support scripts, remove potentially dangerous functions
os.remove = function() end
os.rename = function() end

-- Basic support objects
require "List.lua"
require "Table.lua"
require "Timeout.lua"

-- Constants and globals
require "Const.lua"			-- GRAVITY provided
require "Const2.lua"			-- RampageAI-specific
require "Globals.lua"			-- Changable globals

-- Function support scripts
require "AntiPosLag.lua"		-- anti-position lag
require "AutoAlch.lua"			-- auto-alchemist skills
require "Battle.lua"			-- targetting / attacking
require "Commands.lua"			-- command processing
require "ConfCheck.lua"			-- configuration verification
require "Friends.lua"			-- friend recognition
require "InitSupport.lua"		-- initialization support
require "MonSupport.lua"		-- monster support functions
require "NetUsageMonitor.lua"		-- network usage monitoring
require "Personality.lua"		-- personality support
require "Sequencer.lua"			-- sequencing
require "Utils.lua"			-- misc

-- Debug-support scripts
require "Simulator.lua"			-- RO-function simulator

-- Homunculus-specific Scripts
require "Amistr.lua"
require "Filir.lua"
require "Lif.lua"
require "Vanilmirth.lua"

-- Configuration
require "Conf/Config.lua"
require "Conf/Monsters.lua"		-- monster-selection

-- User-specific Scripts
require "Conf/Custom.lua"

--------------------
-- AI start delay --
--------------------

-- This idea was from Fallen~Angel

function AI(id)
	-- Get the tick-count
	local ticks = GetTick()

	-- Check if we've waited any time
	if AIStartedAt == 0 then
		-- We haven't yet, set the begin time
		AIStartedAt = ticks

		-- Do any pre-AI delay initialization
		AI_PreDelayInit()

		-- Log it
		Log(0,"Delaying AI for %d milliseconds...",
			AIStartWait)
	end

	-- Check if we've waited long enough
	if ticks - AIStartedAt >= AIStartWait then
		-- We've waited, never wait again
		AI = AI_track

		-- Call the new AI function
		AI(id)
	end
end

--------------------------
-- Performance Tracking --
--------------------------

function AI_track(id)
	-- Get the starting ticks
	local start = GetTick()

	-- Check if there was a last cycle
	if CycleTime.Tick ~= 0 then
		-- Update our time-between variable
		CycleTime.Between = CycleTime.Between + start - CycleTime.Tick
	end

	-- Call our main AI
	mainAI(id)

	-- Get the ending ticks
	local endtime = GetTick()

	-- Update our variables
	local delta = endtime - start
	CycleTime.Total = CycleTime.Total + delta
	CycleTime.Count = CycleTime.Count + 1
	CycleTime.Tick = endtime
	if delta > CycleTime.Longest then
		CycleTime.Longest = delta
	end
end

--------------------
-- Initialization --
--------------------

--[[
	The PreDelayInit() is meant to initialize only sections of the script
	that require >>absolutely no<< information about the homunculus itself.

	The delay is there to allow the RO client to initialize its own data
	before the AI script begins to poll for information
--]]
function AI_PreDelayInit()
	-- Do the debugging
	if Debug ~= true then
		-- Make the Log function exit immediately
		Log = NoReturn

		-- Don't track performance
		AI_track = mainAI
	else
		-- Capture the TraceAI function
		RAI_TraceAI = TraceAI
		TraceAI = function(str)
			Log(0,"(TraceAI) %s",tostring(str))
		end

		-- Check the debug level
		if type(DebugLevel) ~= "number" then
			DebugLevel = 0
		end

		-- Is the traceai.txt file screwy?
		if SingleLineLogs == true then
			Log = LogWorkaround
		else
			-- If its not screwy, its going to append to the previous AI

			-- Lets make it easier to distinguish
			Log(0,"\r\n\r\n\r\nAI reset.\r\n")
		end

		-- Check our DebugMark option
		if type(DebugMark) ~= "number" then
			DebugMark = 20000
		elseif DebugMark < 2000 then
			DebugMark = 2000
		end

		-- Setup a timeout for logging every X seconds
		if DebugLevel >= 2 then
			-- Check if we want to track network usage generated by the homunculus script
			local debugMarkFunc
			if EnableNetUsageMonitor ~= true then
				-- We don't, so set the function to not use it
				debugMarkFunc = function(ticks)
					Log(2,"\r\n\t-- mark (%d ms since last, %d ms longest cycle, " ..
						"avg %f ms cycle, avg %f ms between) --",
						ticks,CycleTime.Longest,CycleTime.Total / CycleTime.Count,
						CycleTime.Between / CycleTime.Count)
					CycleTime.Total = 0
					CycleTime.Count = 0
					CycleTime.Between = 0
					CycleTime.Longest = 0
				end
			else
				-- Set the function to display network usage caused by homunculus
				debugMarkFunc = function(ticks)
					Log(2,"\r\n\t-- mark (%d ms since last, %d ms longest cycle, " ..
						"avg %f ms cycle, avg %f ms between, ~%d packets sent) --",
						ticks,CycleTime.Longest,CycleTime.Total / CycleTime.Count,
						CycleTime.Between / CycleTime.Count, NetworkPacketsSent)
					CycleTime.Total = 0
					CycleTime.Count = 0
					CycleTime.Between = 0
					CycleTime.Longest = 0
					NetworkPacketsSent = 0
				end
			end
	
			-- Add the mark function to our timeouts
			Timeout.add(DebugMark, debugMarkFunc, true)
		else
			-- Since we don't care, replace the AI() function
			replaceAI = true
		end
	end

	-- Debug
	Log(0,"Initializing homunculus AI...")

	-- Output our ScriptLocation
	if not ScriptLocation then
		ScriptLocation = ""
	end
	Log(0,"Script seems to be located at %q.",ScriptLocation)

	-- Check our configuration options
	CheckConfig()

	-- Setup basic handling
	SpecificAI	= NoReturn			-- no AI function
	SEQ_Custom	= FalseReturn			-- no custom sequences
	CMD_CustAdvMove	= FalseReturn			-- no custom advanced move commands

	ProcessCommand	= AI_Command			-- Process basic GetMsg()s
	GetHomunMotion	= AI_HomunMotion		-- Get the homunculus's motion

	-- Setup our check to auto-use a homunculus's healing/support skills
	-- 				(this is Homunculus specific)
	HomunSkillCheck = FalseReturn

	-- Seed the random number generator
	math.randomseed(GetTick())
	Log(1,"First random number from this seed: %q",tostring(math.random()))

	-- See if the homunculus should try to stay ahead of the alchemist
	if HomunRunAhead then
		-- Set the function to check the distance
		CheckDistance = AI_CheckDistance_Lead
	else
		-- Set the function to check the distance
		CheckDistance = AI_CheckDistance_Follow
	end

	-- See if we should default to agressive
	if DefaultAggro then
		Mode = MODE_AGGRO
	end

	-- Call our modular initializations
	InitAPIFuncs()
	InitNetworkUsageMonitor()
	InitAntiPosLag()
	InitAttackLog()
	InitAcquire()
	InitKillSteal()
	InitFriends()
	InitHPDeltas()
	InitAmountOrPercent()
	InitManualSkills()
	InitOwnerSkills()
	InitPersonality()

	-- Check if we don't want to ignore
	if ActorIgnoreTime > 0 then
		-- Initialize a timer to check the list of ignored monsters every half-second
		Timeout.add(500,
			function(ticks)
				-- Call the ignored monster check
				CheckIgnored(ticks)
			end,
			-- Repeatable = true
			true)
	else
		-- Replace the function that adds new ignores
		AddIgnore = NoReturn
	end

	-- Convert our MonsterTypeDB and MonsterSpecDB tables to be faster
	ConvertMDB()

	-- Initialize Advanced Movement Commands
	AMC_Init()
end

function mainAI(id)
	-- Log it
	Log(0,"Commencing post-AI-delay initialization...")

	-- Get us a temporary flag to see if we should replace AI() or mainAI()
	local replaceAI = false

	-- Default the next AI function to be AI_part2
	local nextAI = AI_part2

	-- Check if we're not debugging
	if Debug ~= true or type(DebugMark) ~= "number" or DebugMark < 1 then
		-- Replace the performance tracking
		replaceAI = true
	end

	-- Determine our owner's ID
	OwnerID = GetV(V_OWNER,id)
	Log(0,"Owner ID: %d", OwnerID)

	-- Determine information about our homunculus
	HomunID = id
	HomunRange = GetV(V_ATTACKRANGE,HomunID)
	HomunSpecificType = GetV(V_HOMUNTYPE,HomunID)

	-- Output info
	Log(0,"Homunculus ID: %d", HomunID)
	Log(0,"Homunculus Attack Range: %2.1f", HomunRange)

	-- Determine if the homunculus is advanced
	local advancedStr = ""
	if HomunType == AMISTR_H or HomunType == AMISTR_H2 or
		HomunType == FILIR_H or HomunType == FILIR_H2 or
		HomunType == LIF_H or HomunType == LIF_H2 or
		HomunType == VANILMIRTH_H or HomunType == VANILMIRTH_H2 then

		-- Its an advanced homunculus
		HomunAdvancedType = true

		-- Set the advanced string
		advancedStr = " High"

			-- eg, "Homunculus Type: High Vanilmirth"
			-- or, "Homunculus Type: Vanilmirth"
	end

	-- Determine the basic type of homunculus
	HomunType = HomunSpecificType
	if HomunType == AMISTR2 or HomunType == AMISTR_H or
		HomunType == AMISTR_H2 then

		HomunType = AMISTR

	elseif HomunType == FILIR2 or HomunType == FILIR_H or
		HomunType == FILIR_H2 then

		HomunType = FILIR

	elseif HomunType == LIF2 or HomunType == LIF_H or
		HomunType == LIF_H2 then

		HomunType = LIF

	elseif HomunType == VANILMIRTH2 or HomunType == VANILMIRTH_H or
		HomunType == VANILMIRTH_H2 then

		HomunType = VANILMIRTH
	end

	-- Initialize the script based on which homunculus we have
	if HomunType == AMISTR then
		Log(0,"Homunculus Type:" .. advancedStr .. " AMISTR")
		Amistr_Init()
	elseif HomunType == FILIR then
		Log(0,"Homunculus Type:" .. advancedStr .. " FILIR")
		Filir_Init()
	elseif HomunType == LIF then
		Log(0,"Homunculus Type:" .. advancedStr .. " LIF")
		Lif_Init()
	elseif HomunType == VANILMIRTH then
		Log(0,"Homunculus Type:" .. advancedStr .. " VANILMIRTH")
		Vanilmirth_Init()
	end

	-- Initialize the homunculus's skills
	InitHomunSkills()

	-- Initialize any customizations
	InitCustom()

	-- Check if we should assume we're in a PVP zone
	if AssumePVP then
		-- Assume we're in a pvp zone all the time
		PVPZone = true

		-- Initialize the PVP-zone specific custom addons
		InitCustomPVP()
	elseif AssumeWoE then
		-- Assume we're in a woe zone all the time
		PVPZone = true
		WoEZone = true

		-- Initialize the WoE-zone specific custom addons
		InitCustomWoE()
	elseif CheckPVP then
		-- Log it
		Log(0,"Auto-detecting PVP zone...")

		-- Set the next function to be a PVP-auto-detector
		nextAI = AI_pvp_check
	end

	-- Check if we should replace the AI() function (so we don't check performance)
	if replaceAI then
		AI = nextAI
	end

	-- We're done initializing, never do it again
	mainAI = nextAI

	-- Now that we're done initializing, call the main logic
	mainAI()
end

------------------------
-- PVP Auto-detection --
------------------------

--[[

	NOTE: Go figure. As soon as I implement auto-PVP detection, GRAVITY
		makes it so that homunculi are no longer able to attack
		themselves.

--]]

function AI_pvp_check()
	-- Get the current ticks
	local ticks = GetTick()

	-- Get the current amount of HP/SP that we have
	local hp = GetV(V_HP,HomunID)
	local sp = GetV(V_SP,HomunID)

	-- Check if we count as a monster
	if RAI_IsMonster(HomunID) == 1 then
		-- Log it
		Log(0,"PVP zone detected!")

		-- We're in a PVP zone
		PVPZone = true

		-- Initialize customizations
		InitCustomPVP()

	-- Check if we're attacking ourself or we have >less< SP than we did
	--	before the PVP check
	elseif GetV(V_TARGET,HomunID) == HomunID or
		(PrePVPCheckSP ~= nil and sp < PrePVPCheckSP) then

	--[[
		NOTE: If we did NOT show up as a monster, but we CAN attack ourself,
			then we must be in a War of Emperium zone.
	--]]

		-- Default to being in a WoE zone
		local foundWoE = true

		-- Check if we used SP
		if PrePVPCheckSP and sp < PrePVPCheckSP then
			-- Check if our HP is lower because of it
			if hp >= PrePVPCheckHP then
				-- WoE zone not detected. SP was used, but we didn't do any damage
				foundWoE = false
			end
		end

		-- Check if we still found WoE
		if foundWoE then
			-- Log it
			Log(0,"WoE zone detected!")
	
			-- We're in a PVP zone
			PVPZone = true
			WoEZone = true
	
			-- Initialize customizations
			InitCustomWoE()
		end
	end

	-- Check how long its been since we started
	if AIStartedAt + AIStartWait + 500 < ticks or PVPZone then
		-- We've waited long enough, set the next AI part

		-- Check if we haven't detected a PVP zone
		if not PVPZone then
			-- Log it
			Log(0,"PVP zone >NOT< detected!")
		end

		-- Check to see if the main AI is us
		if AI == AI_pvp_check then
			-- Set the next AI
			AI = AI_part2
		end

		-- Set the main AI
		mainAI = AI_part2

		-- Call the next AI
		return mainAI()
	end

	-- Default to not checking skills
	local skillCheck = false

	-- Check type of homunculus
	if HomunType == FILIR then
		-- Check if we have enough sp
		if sp >= 4 then
			-- Check WoE zone by using moonlight
			skillCheck = true
		end
	elseif HomunType == VANILMIRTH then
		-- Check if we have enough sp
		if sp >= 22 then
			-- Check WoE zone by using caprice
			skillCheck = true
		end
	end

	-- Check if we want to check skills
	if skillCheck then
		-- Save our HP/SP values
		PrePVPCheckSP = sp
		PrePVPCheckHP = hp

		-- Send a command to attack ourself
		SkillObject(HomunID,1,8009,HomunID)	-- moonlight
		SkillObject(HomunID,1,8013,HomunID)	-- caprice
	end

	-- Send a command to attack ourself
	Attack(HomunID,HomunID)
end

-------------------
-- Main AI Logic --
-------------------

function AI_part2()
	-- Get our owner's motion (to be used later)
	OwnerMotion = GetOwnerMotion()

	-- Get our homunculus's motion
	HomunMotion = GetHomunMotion()

	-- Get the command from the user interface
	local shift = false
	local msg = GetMsg(HomunID)

	-- If there is no message, check for a "reserved" message
	if msg[1] == NONE_CMD then
		shift = true
		msg = GetResMsg(HomunID)
	end

	-- Process the command
	if msg[1] ~= NONE_CMD then
		ProcessCommand(msg,shift)
	end

	-- Check the HP deltas
	CalculateHPDeltasAlch()
	CalculateHPDeltasHomu()

	-- Gather information about each actor
	GetActorInfo()

	-- Check if its time that we try use a skill
	PauseAttackSkill = HomunSkillCheck()

	-- Process our timeouts
	Timeout.iterate()

	-- Process the sequence and grab the next one thats scheduled
	local seq = ProcessSequence()

	-- Check if we're doing anything
	local emptySeq = false
 	if seq[1] == SEQ_NONE then
		-- Not doing anything, so we'll process sequence again at
		--	the end (in case we do want to start doing something
		--	this cycle)
		emptySeq = true
	end

	-- Check if we should start chasing our owner (xhint and yhint)
	local chase,distmult,xh,yh = CheckDistance(seq)
	if chase then
		-- If we're moving...
		if seq[1] == SEQ_MOVE or seq[1] == SEQ_MOVE_TO or
			seq[1] == SEQ_MOVE_TO_POS or seq[1] == SEQ_MOVE_TO_NEG or
			seq[1] == SEQ_PREROUTED then

			-- Then stop. (wasted command)
			--Move(HomunID,GetV(V_POSITION,HomunID))

			-- And remove it from the stack too
			List.popleft(AI_seq)
		else
			-- Make a new sequence
			seq = { }
		end

		-- Reuse the seq we have
		seq[1] = SEQ_MOVE_TO
		seq[2] = OwnerID
		seq[3] = HomunFollowDistance * distmult
		seq[4] = xh
		seq[5] = yh
		seq.loglevel = 3

		-- Call the sequencer's "move-to-actor" function
		local dummy
		dummy,seq = SEQ_AMove(seq)

		-- If the function wishes to be called again,
		-- push it onto the front of our AI_seq stack
		if seq ~= nil then
			List.pushleft(AI_seq,seq)
		end

		-- Do nothing else now (disabled)
		--return
	end

	-- Get the current tick count
	local ticks = GetTick()

	-- Acquire a target
	if ticks > AcquireTargetNextTime then
		-- Acquire a target for the homunculus
		HomuAcquireTarget()

		-- Acquire a target for the alchemist
		AlchAcquireTarget()

		-- Reset the timeout
		AcquireTargetNextTime = ticks + AcquireTargetTimeout
	end

	-- Check auto-alchemist skills
	CheckAlchSkills()

	-- Check for position lag
	CheckPosLag()

	-- Call customized AI
	CustomAI()

	-- Check if we're doing anything
	if List.size(AI_seq) < 1 then
		-- Check if we should be considered idle now
		if HomunLast.Sequence + TimeBeforeIdle < ticks then
			-- Do whatever the personality says to do while idle
			Personality_Idle()
		end
	else
		-- Set the last sequence
		HomunLast.Sequence = ticks
	end

	-- Check if our sequencer was empty originally, but we have something now
	if List.size(AI_seq) > 0 and emptySeq then
		-- Process the sequence again, to get started immediately
		ProcessSequence()
	end

	-- Save some information
	OwnerLast.Motion = OwnerMotion
	OwnerLast.X,OwnerLast.Y = GetV(V_POSITION,OwnerID)
end

-------------------------------
-- GetHomunMotion() function --
-------------------------------

function AI_HomunMotion()
	-- Get the motion
	local motion = GetV(V_MOTION,HomunID)

	--[[ Homunculus is never counted as "attacking" for some reason

	-- Check if the motion is an attack
	if motion == MOTION_ATTACK or motion == MOTION_ATTACK2 then
		-- Set the last time it attacked
		HomunLast.Attack = GetTick()
	end

	--]]

	return motion
end

--------------------------------
-- GetOwnerMotion() functions --
--------------------------------

function AI_GetOwnerMotion_Normal()
	local motion = GetV(V_MOTION,OwnerID)

	-- Check if the motion is an attack
	if motion == MOTION_ATTACK or motion == MOTION_ATTACK2 then
		-- Set the last time they attacked
		OwnerLast.Attack = GetTick()
	end

	-- Check if the motion is a movement
	if motion == MOTION_MOVE then
		OwnerLast.Move = GetTick()
	end

	-- Check if the motion is a pickup
	if motion == MOTION_PICKUP then
		OwnerLast.Pickup = GetTick()
	end

	-- Check if the motion is a sit
	if motion == MOTION_SIT then
		-- Check if we weren't already sitting
		if OwnerLast.Motion ~= MOTION_SIT then
			OwnerLast.Sit = GetTick()
		end
	end

	-- Check if they're dead (sometimes RO bugs and leaves homun out)
	if motion == MOTION_DEAD then
		-- Pretend like they're sitting -- no skill usage
		motion = MOTION_SIT
	end

	return motion
end

function AI_GetOwnerMotion_NoPhen()
	-- Get their motion
	local motion = AI_GetOwnerMotion_Normal()

	-- Check if they got hit
	if motion == MOTION_DAMAGE then
		-- Reset the casting time, we got hit
		OwnerSkillTimeout = 0

		-- Return the motion
		return motion
	end

	-- Reuse the phen function
	return AI_GetOwnerMotion_Phen()
end

function AI_GetOwnerMotion_Phen()
	-- Get the motion
	local motion = AI_GetOwnerMotion_Normal()

	-- Check if we're casting
	if OwnerSkillTimeout ~= 0 then
		-- Check if its expired
		if OwnerSkillTimeout <= GetTick() then
			-- Its expired
			OwnerSkillTimeout = 0
		else
			-- Not expired, they're casting
			motion = MOTION_CASTING
		end
	end

	-- Return their motion
	return motion
end

---------------------------
-- Follow/Lead Functions --
---------------------------

-- Function to decide if the homunculus is already chasing its owner
function AI_CheckDistance_PreTest(seq)
	-- Default to no
	local ret = false

	-- Check if the sequence is a move-to our owner
	if (seq[1] == SEQ_MOVE_TO or seq[1] == SEQ_MOVE_TO_POS or
		seq[1] == SEQ_MOVE_TO_NEG) and seq[2] == OwnerID then
		-- We are
		return true
	end

	return false
end

-- Function to decide if a homunculus should start chasing its master (follow)
function AI_CheckDistance_Follow(seq)
	-- Default to not checking the distance
	local ret = false

	-- Check if we're following our owner already
	if AI_CheckDistance_PreTest(seq) then
		-- No reason to start following again
		return false
	end

	-- Get the distance
	local dist
	if seq[1] == SEQ_MOVE then
		-- Get the block distance between our target location and owner
		dist = APBlocks(OwnerID,seq[2],seq[3])

	elseif seq[1] == SEQ_MOVE_TO or seq[1] == SEQ_MOVE_TO_POS or
		seq[1] == SEQ_MOVE_TO_NEG or seq[1] == SEQ_PREROUTED then

		-- Get the block distance between our target and owner
		dist = AABlocks(seq[2],OwnerID)

		-- Check if its just following them
		if seq[1] == SEQ_MOVE_TO_POS and dist <= SightRange then
			dist = dist - seq[3]
		end

	elseif seq[1] == SEQ_ATTACK then
		-- Get the block distance between our target monster and owner
		dist = AABlocks(Target.ID,OwnerID)

		-- Check for a special case
		if dist > MaxHomunRange and dist <= SightRange then
			-- Its not pursuing them, don't run to owner
			return false,1
		end

	elseif seq[1] == SEQ_KITE then
		-- Get the block distance between our owner and monster
		dist = AABlocks(Target.ID,OwnerID)
	else
		-- Get the block distance between our homun and owner
		dist = AABlocks(HomunID,OwnerID)
	end

	-- Range to our owner
	local range = MaxHomunRange

	-- Check if our owner is moving
	if OwnerMotion == MOTION_MOVE then
		-- Owner is moving

		-- Check if our owner is moving away from us/our target
		if dist > OwnerLast.Dist then
			-- Shorten the range
			range = range - HomunRangeChange
		end

		-- Check if we were told to sit a a specific place
		if HoldPosition == true and dist > range then
			-- Clear the hold position flag
			HoldPosition = false
		end
	end

	-- Check if we should start chasing our owner
	if HoldPosition == false and dist > range then
		-- We're going to have to run to our owner
		ret = true
	end

	-- Save the owner's distance for this cycle
	OwnerLast.Dist = dist

	--
	return ret,1
end

-- Function to decide if a homunculus should run ahead of its master (lead)
function AI_CheckDistance_Lead(seq)
	-- Check if we're already leading
	if AI_CheckDistance_PreTest(seq) then
		-- Don't do another lead command
		return false
	end

	-- Get old distance
	local olddist = OwnerLast.Dist

	-- First and foremost, check if we need to follow
	if AI_CheckDistance_Follow(seq) then
		return true,1
	end

	-- Get the owner's old/new X and Y
	local x,y = GetV(V_POSITION,OwnerID)

	-- Check if we're holding position
	if HoldPosition == true then
		-- Don't chase our owner
		return false
	end

	-- Check if we're in the middle of fighting
	if seq[1] == SEQ_ATTACK then
		-- Don't chase our owner
		return false
	end

	-- Check if our owner has stood still
	if OwnerLast.X == 0 or (OwnerLast.X == x and OwnerLast.Y == y) then
		-- Don't chase, they're standing still
		return false
	end

	-- Regular with hints
	return true,-1,OwnerLast.X,OwnerLast.Y
end

-----------------------
-- Simulator Support --
-----------------------

--[[ Check if we're in simulator mode (read Simulator.lua for more information)

	NOTE: Lua is executed top-to-bottom. When RO needs to start the
		homunculus script, it executes AI.lua (which in turn executes
		the others with "require" statements) and then RO continues to
		run the function AI() repeatedly. Since I have written
		functions to help me test the script versus certain
		situations, I use this function call to determine if we are
		are actually in the RO client or not.

	NOTE: If we are not in the RO client, CheckSimulation() will also run
		the script against the test case defined by Simulator.lua
--]]
CheckSimulation()
