--[[

This file is part of RampageAI.

RampageAI is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

RampageAI is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RampageAI; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

--]]

------------------------
-- Amistr-specific AI --
------------------------

--[[
	Unsupported Skills:

	Bloodlust	( ID 8008, SP 120, effect duration (1,3,5)min, delay (5,10,15)min )
				*info from Forgotten Dreams of iRO Loki*
--]]

-- SP required, and lasting time for each level of Amistr Bulwark
BulwarkSP = { 20, 25, 30, 35, 40 }
BulwarkTimes = { 40, 35, 30, 25, 20 }

-- When Amistr's skill delay is expected to wear off
AmistrTimeout = 0

-- Information for when we cast Amistr Bulwark
PreBulwarkSP = 0
BulwarkLevelTried = 0

-- Initialization function
function Amistr_Init()
	-- Check if support skills are even allowed
	if SupportSkills then
		-- Check both our Amistr Bulwark options
		if AmistrBulwarkHomun < 1 then
			AmistrBulwarkHomun = 1000
		end
		if AmistrBulwarkOwner < 1 then
			AmistrBulwarkOwner = 1000
		end
		
		-- Check both our castling options
		if CastlingAttackers < 1 then
			CastlingAttackers = 1000
		end

		-- Set the HomunSkillCheck() function
		HomunSkillCheck = Amistr_SkillCheck
	end
end

-- Amistr's skill-check function (called every cycle)
function Amistr_SkillCheck(seq)
	-- Get our usable SP amount
	local sp = SPUsable(HomunID)

	-- Get the time in ticks
	local ticks = GetTick()

	-- Check if we saved the SP amount
	if PreBulwarkSP ~= 0 then
		-- We have, check if its been cast

		if sp < PreBulwarkSP then
			-- We have less SP

			-- Calculate how much SP was used
			local spUsed = PreBulwarkSP - sp

			-- Determine which level of Bulwark was used
			local i,levelUsed
			for i=BulwarkLevelTried,1,-1 do
				-- Check if we used enough SP for this to be the level
				if BulwarkSP[i] == spUsed then
					-- Yep, we'll assume its the level used then
					levelUsed = i
					break
				end
			end

			-- Check levelUsed
			if levelUsed == nil then
				levelUsed = BulwarkLevelTried
			end

			-- Log it
			Log(2,"Successfully cast Amistr Bulwark level %d.",levelUsed)

			-- Set the timeout
			AmistrTimeout = ticks + (BulwarkTimes[levelUsed] * 1000)

			-- Remove the SP amount
			PreBulwarkSP = 0

			-- We shouldn't pause attack skills
			--	(what attack skills? nonetheless, return false)
			return false
		else
			-- We have the same amount of SP

			-- After 3 waits, give it another try
			if math.mod(BulwarkTries,3) == 0 then
				-- Check if we've tried 3 times already
				if (BulwarkTries/3) > 3 then
					-- Log it
					Log(2,"Failed to cast Amistr Bulwark, giving up.")

					-- Reset
					PreBulwarkSP = 0

					-- We aren't trying
					return true
				else
					-- Log it
					Log(2,"Trying again to auto-cast Amistr Bulwark (level %d).", BulwarkLevelTried)

					-- Check if the homunculus is moving
					if HomunMotion == MOTION_MOVE then
						-- The skill usage would stop it
						BypassMove(HomunID,GetV(V_POSITION,HomunID))
					end

					-- Cast Amistr Bulwark
					SkillObject(HomunID, BulwarkLevelTried, 8006, HomunID)
				end
			end

			-- Check if we should increment the tries counter
			if ticks >= BulwarkTriesTicks then
				-- Increment the number of tries
				BulwarkTries = BulwarkTries + 1
				
				-- Set the new tries timeout
				BulwarkTriesTicks = ticks + 50
			end

			-- Still waiting, so return true
			return true
		end
	elseif SkillAllowed(SKILL_TYPE_SUPPORT) then
		-- We haven't, check if we need to

		-- Check if the timeout has expired
		if AmistrTimeout == 0 or ticks >= AmistrTimeout then

			-- Get the targets of our homunculus and owner
			homunEnemies = ActorEnemies(HomunID)
			ownerEnemies = ActorEnemies(OwnerID)

			-- Check if our owner has too many things attacking him,
			-- and is below a certain amount of HP
			if Table.getn(ownerEnemies) >= CastlingAttackers and
				ActorHP(OwnerID) < CastlingHP then
				-- Attempt to cast Castling
				
				-- Log it
				Log(2,"Attempting to auto-cast Castling, to save our owner.")
				
				-- Cast castling
				SkillObject(HomunID, 5, 8005, HomunID)
				
				-- Set the amistr timeout to delay skill usage
				AmistrTimeout = ticks + CastlingDelay
				
				-- We're done for now
				return true
			end

			-- Check if we have too many things targetting us
			if Table.getn(homunEnemies) >= AmistrBulwarkHomun or
				Table.getn(ownerEnemies) >= AmistrBulwarkOwner then
				-- Too many enemies

				-- Determine the level to use
				local i,level
				for i=5,AmistrBulwarkMinLevel,-1 do
					-- Check if we have enough sp
					if sp >= BulwarkSP[i] then
						-- Set the level
						level = i
						break
					end
				end

				-- Make sure we got a level to use
				if level ~= nil then
					-- Log it
					Log(2,"Attempting to auto-cast Amistr Bulwark (level %d), due to aggressives.", level)

					-- Check if the homunculus is moving
					if HomunMotion == MOTION_MOVE then
						-- The skill usage would stop it
						BypassMove(HomunID,GetV(V_POSITION,HomunID))
					end

					-- Cast Amistr Bulwark
					SkillObject(HomunID, level, 8006, HomunID)

					-- Save our SP amount, and level attempted
					PreBulwarkSP = sp
					BulwarkLevelTried = level
					AmistrTimeout = 0
					BulwarkTries = 1
					BulwarkTriesTicks = ticks + 50
				end

				-- We're trying to use a skill
				return true
			end
		end
	end

	-- Probably not trying to use a skill
	return false
end