--[[

This file is part of RampageAI.

RampageAI is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

RampageAI is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RampageAI; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

--]]

-----------------------------------------
-- Initialize Network Usage Monitoring --
-----------------------------------------

function InitNetworkUsageMonitor()
	-- Check if they want to monitor network usage
	if EnableNetUsageMonitor == true then
		-- They want to track the outgoing traffic

		-- First, copy all of the existing functions we're going to replace
		PreTrack_Move = Move
		PreTrack_MoveToOwner = MoveToOwner
		PreTrack_Attack = Attack
		PreTrack_SkillObject = SkillObject
		PreTrack_SkillGround = SkillGround

		-- Now replace the functions with our own (that will
		--	only count the packets)
		Move = Track_Move
		MoveToOwner = Track_MoveToOwner
		Attack = Track_Attack
		SkillObject = Track_SkillObject
		SkillGround = Track_SkillGround

		-- Never do this again
		InitNetworkUsageMonitor = NoReturn
	end
end

---------------------------
-- Replacement Functions --
---------------------------

function Track_Move(id1,x,y)
	-- Increment the number of packets sent
	NetworkPacketsSent = NetworkPacketsSent + 1

	-- Call the original function
	PreTrack_Move(id1,x,y)
end

function Track_MoveToOwner(id)
	-- Increment the number of packets sent
	NetworkPacketsSent = NetworkPacketsSent + 1

	-- Call the original function
	PreTrack_MoveToOwner(id)
end

function Track_Attack(id1,id2)
	-- Increment the number of packets sent
	NetworkPacketsSent = NetworkPacketsSent + 1

	-- Call the original function
	PreTrack_Attack(id1,id2)
end

function Track_SkillObject(fromid,level,skillid,toid)
	-- Increment the number of packets sent
	NetworkPacketsSent = NetworkPacketsSent + 1

	-- Call the original function
	PreTrack_SkillObject(fromid,level,skillid,toid)
end

function Track_SkillGround(fromid,level,skillid,x,y)
	-- Increment the number of packets sent
	NetworkPacketsSent = NetworkPacketsSent + 1

	-- Call the original function
	PreTrack_SkillGround(fromid,level,skillid,x,y)
end
